"""
Defines a few useful functions and most importantly, the low level
hdf5 reader and writer. It also serves as a placeholder for variables
that can be changed via user functions.
"""
import os
import warnings
import numpy as np
import h5py
from functools import wraps
from . import settings
from . import units as pu
from .cython.get_index_of_region import get_cube, get_radial_range
from .cython.get_index_of_region import get_cube_plus_thin_layer
from .cython.get_index_of_region import get_radial_range_plus_thin_layer
from .cython.get_index_of_region import get_rotated_cube
from .cython.get_index_of_region import get_rotated_cube_plus_thin_layer
from .cython.openmp_info import simple_reduction, get_openmp_settings

# These will be set by the user using the add_user_unit function
user_unit_dict = {'default': {},
                  'voronoi_cells': {},
                  'dark_matter': {},
                  'stars': {},
                  'black_holes': {},
                  'groups': {},
                  'subhalos': {}}


def get_project_root_dir():
    """
    Returns the root directory of the local Paicos copy.
    """
    path = os.path.dirname(os.path.abspath(__file__))
    r_dir = ''
    path_split = path.split('/')
    for ii in range(1, len(path_split) - 1):
        r_dir += '/' + path_split[ii]

    return r_dir + '/'


# Set the root_dir for Paicos
root_dir = get_project_root_dir()


def _split_filename(filenamewithpath):
    """
    A convenience function for getting the basedir, basename and snapsnum
    from a filename including its path.

    Example usage:
        print(split_filename('data/small_non_comoving_007.hdf5'))
        print(split_filename('small_non_comoving.hdf5'))
        print(split_filename('small_non_comoving.0.hdf5'))

    """
    err_msg = "needs to be a hdf5 file, .e.g. filenamewithpath = 'data/snap_010.hdf5'"

    assert filenamewithpath[-5:] == '.hdf5', err_msg

    snapnum = None

    index = filenamewithpath[::-1].find('_')
    if index != -1:
        underscore_loc = -index - 1

    if filenamewithpath[underscore_loc] == '_':
        snapnum_str = filenamewithpath[underscore_loc + 1:underscore_loc + 4]
        if np.char.isdigit(snapnum_str):
            snapnum = int(snapnum_str)

    filename = os.path.basename(filenamewithpath)
    basedir = filenamewithpath.replace(filename, '')

    if snapnum is None:
        basename = filename[:-5]
    else:
        basename = filename[:underscore_loc]

    return basedir, basename, snapnum


def save_dataset(hdf5file, name, data=None, data_attrs={},
                 group=None, group_attrs={}):
    """
    Create dataset in *open* hdf5file ( hdf5file = h5py.File(filename, 'w') )
    If the data has units then they are saved as an attribute.
    """
    assert isinstance(data_attrs, dict)
    data_attrs = dict(data_attrs)
    assert isinstance(group_attrs, dict)
    group_attrs = dict(group_attrs)

    # Allow for storing data sets in groups or nested groups
    if group is None:
        path = hdf5file
    else:
        if group not in hdf5file:
            hdf5file.create_group(group)
            for key in group_attrs:
                hdf5file[group].attrs[key] = group_attrs[key]
        path = hdf5file[group]

    # Save data set
    if hasattr(data, 'unit'):
        path.create_dataset(name, data=data.value)
    else:
        path.create_dataset(name, data=data)

    # Write attributes
    if isinstance(data, pu.PaicosTimeSeries):
        data_attrs.update(data.hdf5_attrs)
    elif hasattr(data, 'unit'):
        data_attrs['unit'] = data.unit.to_string()
    for key in data_attrs.keys():
        path[name].attrs[key] = data_attrs[key]

    # Check if scale_factors are already saved
    if isinstance(data, pu.PaicosTimeSeries):
        if 'scale_factor' not in path.keys():
            path.create_dataset('scale_factor', data=data.a)
        else:
            np.testing.assert_array_equal(path['scale_factor'][...], data.a)


def load_dataset(hdf5file, name, group=None):
    """
    Load dataset, returning a paicos quantity if the attributes
    contain units and units are enabled.
    """

    if not isinstance(hdf5file, h5py.File):
        if isinstance(hdf5file, str):
            hdf5file = h5py.File(hdf5file, 'r')
        else:
            msg = "'util.load_dataset' needs a file name or an open hdf5 file"
            raise RuntimeError(msg)

    comoving_sim = bool(hdf5file['Parameters'].attrs['ComovingIntegrationOn'])
    time = hdf5file['Header'].attrs['Time']
    hubble_param = hdf5file['Parameters'].attrs['HubbleParam']
    if hubble_param == 0.:
        hubble_param = 1.0

    # Allow for loading data sets in groups or nested groups
    if group is None:
        path = hdf5file
    else:
        if group not in hdf5file:
            hdf5file.create_group(group)
        path = hdf5file[group]

    data = path[name][()]

    if settings.use_units:
        if 'unit' in path[name].attrs.keys():
            unit = path[name].attrs['unit']
            if 'Paicos' in path[name].attrs.keys():
                if path[name].attrs['Paicos'] == 'PaicosTimeSeries':
                    if comoving_sim:
                        time = path['scale_factor'][...]
                    else:
                        time = path['time'][...]
                    data = pu.PaicosTimeSeries(data, unit, a=time, h=hubble_param,
                                               comoving_sim=comoving_sim)
                elif path[name].attrs['Paicos'] == 'PaicosQuantity':
                    data = pu.PaicosQuantity(data, unit, a=time, h=hubble_param,
                                             comoving_sim=comoving_sim)
            else:
                data = pu.PaicosQuantity(data, unit, a=time, h=hubble_param,
                                         comoving_sim=comoving_sim)
    return data


def remove_astro_units(func):
    """
    This is a decorator function that takes in a function and returns a new
    function that removes any astro units from the function's arguments and
    keyword arguments before calling the original function. The decorator
    checks if the argument or keyword argument has a 'unit' attribute and, if
    so, replaces it with its 'value' attribute. This is useful for functions
    that do not support astro units or need to work with raw values.
    """
    @wraps(func)
    def remove_astro_units_inner(*args, **kwargs):

        unit_dict = {'length': []}
        # Create new args
        new_args = list(args)
        for ii, new_arg in enumerate(new_args):
            if hasattr(new_arg, 'unit'):
                new_args[ii] = new_arg.value
                if isinstance(new_arg, pu.PaicosQuantity):
                    phys_type = new_arg.uq.to_physical.unit.physical_type
                    unit = new_arg.uq.unit
                else:
                    phys_type = new_arg.unit.physical_type
                    unit = new_arg.unit
                if str(phys_type) == 'length':
                    unit_dict['length'].append(unit)

        for ii in range(len(unit_dict['length'])):
            if unit_dict['length'][ii] != unit_dict['length'][0]:
                err_msg = ("Paicos: Not all quantities with physical type "
                           + "'length' have the same units. Stopping here "
                           + "to prevent unit-related errors.")
                print("Contents of unit_dict['length']", unit_dict['length'])
                raise RuntimeError(err_msg)

        # Create new kwargs
        new_kwargs = kwargs  # dict(kwargs)
        for key, kwarg in kwargs.items():
            if hasattr(kwarg, 'unit'):
                new_kwargs[key] = kwarg.value

        return func(*new_args, **new_kwargs)
    return remove_astro_units_inner


@remove_astro_units
def get_index_of_radial_range(pos, center, r_min, r_max):
    """
    Get a boolean array of positions, pos, which are inside the spherical
    shell with inner radius r_min and outer radius r_max, centered at center.
    """
    assert center.shape[0] == 3
    x_c, y_c, z_c = center[0], center[1], center[2]
    index = get_radial_range(pos, x_c, y_c, z_c, r_min, r_max,
                             settings.numthreads)
    return index


@remove_astro_units
def get_index_of_radial_range_plus_thin_layer(pos, center, r_min, r_max, thickness):
    """
    Get a boolean array for the positions, pos, which are inside the spherical
    shell with inner radius r_min and outer radius r_max, centered at center.
    The thickness array argument adds a variable thickness, e.g., the cell
    diameters, so that point close to the selection are also included.
    """
    assert center.shape[0] == 3
    x_c, y_c, z_c = center[0], center[1], center[2]
    index = get_radial_range_plus_thin_layer(pos, x_c, y_c, z_c, r_min, r_max,
                                             thickness, settings.numthreads)
    return index


@remove_astro_units
def get_index_of_cubic_region(pos, center, widths, box):
    """
    Get a boolean array to the position array, pos, which are inside a cubic
    region.

    Parameters
    ----------
    pos : array
         position array with dimensions = (n, 3)
    center : array with length 3
             The center of the box (x, y, z).
    widths : array with length 3
             The widths of the box.
    box : float
                 The box size of the simulation (e.g. snap.box).
    """
    x_c, y_c, z_c = center[0], center[1], center[2]
    width_x, width_y, width_z = widths
    index = get_cube(pos, x_c, y_c, z_c, width_x, width_y, width_z, box,
                     settings.numthreads)
    return index


@remove_astro_units
def get_index_of_cubic_region_plus_thin_layer(pos, center, widths, thickness, box):
    """
    Get a boolean array to the position array, pos, which are inside a cubic
    region plus a thin layer with a cell-dependent thickness

    Parameters
    ----------
    pos : array
         position array with dimensions = (n, 3)
    center : array with length 3
             The center of the box (x, y, z).
    widths : array with length 3
             The widths of the box.
    thickness : array
                Array with same length as the position array.
    box : float
         The box size of the simulation (e.g. snap.box).
    """
    x_c, y_c, z_c = center[0], center[1], center[2]
    width_x, width_y, width_z = widths
    index = get_cube_plus_thin_layer(pos, x_c, y_c, z_c, width_x, width_y,
                                     width_z, thickness, box, settings.numthreads)
    return index


@remove_astro_units
def get_index_of_rotated_cubic_region(pos, center, widths, box, orientation):
    """
    Get a boolean array to the position array, pos, which are inside a cubic
    region

    Parameters
    ----------
    pos : array
         position array with dimensions = (n, 3)
    center : array with length 3
             The center of the box (x, y, z).
    widths : array with length 3
             The widths of the box.
    thickness : array
                Array with same length as the position array.
    box : float
         The box size of the simulation (e.g. snap.box).
    """
    x_c, y_c, z_c = center[0], center[1], center[2]
    width_x, width_y, width_z = widths
    unit_vectors = orientation.cartesian_unit_vectors
    index = get_rotated_cube(pos, x_c, y_c, z_c, width_x, width_y,
                             width_z, box,
                             unit_vectors['x'],
                             unit_vectors['y'],
                             unit_vectors['z'],
                             settings.numthreads)
    return index


@remove_astro_units
def get_index_of_rotated_cubic_region_plus_thin_layer(pos, center, widths, thickness,
                                                      box, orientation):
    """
    Get a boolean array to the position array, pos, which are inside a cubic
    region plus a thin layer with a cell-dependent thickness

    Parameters
    ----------
    pos : array
         position array with dimensions = (n, 3)
    center : array with length 3
             The center of the box (x, y, z).
    widths : array with length 3
             The widths of the box.
    thickness : array
                Array with same length as the position array.
    box : float
         The box size of the simulation (e.g. snap.box).
    """
    x_c, y_c, z_c = center[0], center[1], center[2]
    width_x, width_y, width_z = widths
    unit_vectors = orientation.cartesian_unit_vectors
    index = get_rotated_cube_plus_thin_layer(pos, x_c, y_c, z_c, width_x, width_y,
                                             width_z, thickness, box,
                                             unit_vectors['x'],
                                             unit_vectors['y'],
                                             unit_vectors['z'],
                                             settings.numthreads)
    return index


def _check_if_omp_has_issues(verbose=True):
    """
    Check if the parallelization via OpenMP works.
    Parameters
    ----------
    numthreads : int
        Number of threads used in parallelization
    """

    if settings.give_openMP_warnings is False:
        verbose = False

    max_threads = get_openmp_settings(0, False)
    settings.max_threads = max_threads
    if settings.numthreads > max_threads and verbose:
        msg = ('\n\nThe default number of OpenMP threads, {}, '
               + 'exceeds the {} available on your system. Setting '
               + 'numthreads={}. '
               + 'You can set numthreads with e.g. the command\n '
               + 'paicos.set_numthreads(16)\n\n')
        print(msg.format(settings.numthreads, max_threads, max_threads))
        settings.numthreads = max_threads

    n = simple_reduction(1000, settings.numthreads)
    if n == 1000:
        settings.numthreads_reduction = settings.numthreads
        settings.openMP_has_issues = False
    else:
        # We have issues...
        settings.openMP_has_issues = True
        settings.numthreads_reduction = 1

        msg = ("OpenMP seems to have issues with reduction operators "
               + "on your system, so we'll turn it off for those use cases. "
               + "If you're on Mac then the issue is likely a "
               + "compiler problem, discussed here:\n"
               + "https://stackoverflow.com/questions/54776301/"
               + "cython-prange-is-repeating-not-parallelizing.\n\n")
        if verbose:
            warnings.warn(msg)


def _copy_over_snapshot_information(snap, new_filename, mode='r+'):
    """
    Copy over attributes from the original arepo snapshot.
    In this way we will have access to units used, redshift etc
    """
    info_dic = {}
    info_dic['Header'] = dict(snap.Header)
    info_dic['Parameters'] = dict(snap.Parameters)
    info_dic['Config'] = dict(snap.Config)
    with h5py.File(new_filename, mode) as f:
        for group in ['Header', 'Parameters', 'Config']:
            f.create_group(group)
            for key in info_dic[group]:
                f[group].attrs[key] = info_dic[group][key]



from cython.parallel import prange
cimport openmp
import numpy as np
cimport numpy as np

# ctypedef double real_t

ctypedef fused real_t:
    float
    double
    np.float32_t
    np.float64_t

STUFF = "Hi"  # https://stackoverflow.com/questions/8024805/cython-compiled-c-extension-importerror-dynamic-module-does-not-define-init-fu

# https://stackoverflow.com/questions/54776301/cython-prange-is-repeating-not-parallelizing


def get_openmp_settings(mpirank, verbose=True):
    """
    This helper function returns the maximum number of
    openmp threads available on the system.
    """

    cdef int maxthreads = openmp.omp_get_max_threads()
    if verbose:
        print('mpirank {} has maximum of {} threads'.format(mpirank, maxthreads))

    return maxthreads


cpdef double simple_reduction(int n, int num_threads):
    """
    This program is a simple reduction using openmp parallel
    code. The return value should simply be identical to the input, n.
    Used to find a bug cython/openmp bug that at some point occured on MacOs.
    """
    cdef int i
    cdef int mysum = 0

    for i in prange(n, nogil=True, num_threads=num_threads, schedule='static'):
        mysum += 1
    return mysum



"""
Defines the PaicosReader, Histogram2DReader and ImageReader which
can be used to load derived variables.
"""
import os
import h5py
import numpy as np
import traceback
from astropy import units as u
from astropy.cosmology import LambdaCDM
from .. import util
from .. import units as pu
from .. import settings
from ..orientation import Orientation
from ..image_creators.image_creator import ImageCreator


class PaicosReader(dict):
    """
    The PaicosReader can read any hdf5 file that contains the three
    groups: Header, Config and Parameters.

    It uses these to automatically construct a variety quantities
    which become accessible as properties.

    This class is subclassed by the Snapshot and Catalog classes.
    """

    def __init__(self, basedir='.', snapnum=None, basename="snap",
                 basesubdir='snapdir', load_all=True, to_physical=False,
                 verbose=False):

        """
        Initialize the PaicosReader class.

        Parameters
        ----------

            basedir : str
                The path of the directory containing the hdf5 files
                (e.g. the 'output' folder).

            snapnum : int
                e.g. the snapshot number

            basename : str
                name of the file takes the form ``basename_{:03d}.hdf5``
                or ``basename_{:03d}.{}.hdf5``. Default is ``snap``.

            basesubdir : str
                The name of the subfolder. Default is ``snapdir``.

            load_all : bool
                Whether to simply load all data, default is True.

            to_physical : bool
                Whether to convert from comoving to physical
                variables upon load.

            verbose : bool
                Whether to print information, default is False.
        """

        self.to_physical = to_physical
        self.verbose = verbose

        if '.hdf5' in basedir:
            # User is trying to load a single hdf5 file directly, let's see if
            # that can work
            self.filename = str(basedir)
            self.multi_file = False
            basedir, basename, snapnum = util._split_filename(self.filename)
            self.basedir = basedir
            self.snapnum = snapnum
            self.basename = basename
        else:
            self.basedir = basedir
            self.snapnum = snapnum
            self.basename = basename

            self.basesubdir = basesubdir

            # Add a forward slash at the end of basedir if not already present
            if basedir[-1] != '/':
                basedir += '/'

            # If snapnum is None, filename is basedir + basename + '.hdf5'
            if snapnum is None:
                self.filename = basedir + basename + '.hdf5'
                err_msg = f'File: {self.filename} not found'
                if not os.path.exists(self.filename):
                    raise FileNotFoundError(err_msg)
            else:
                # Set filenames for single_file and multi_file cases
                single_file = basename + "_{:03d}.hdf5"
                multi_file = basesubdir + '_{:03d}/' + basename + '_{:03d}.{}.hdf5'
                multi_wo_dir = basename + '_{:03d}.{}.hdf5'

                single_file = basedir + single_file.format(snapnum)
                multi_file = basedir + multi_file.format(snapnum, snapnum, '{}')
                multi_wo_dir = basedir + multi_wo_dir.format(snapnum, '{}')

                # Check if single_file exists
                if os.path.exists(single_file):
                    self.multi_file = False
                    self.filename = single_file
                elif os.path.exists(multi_file.format(0)):
                    self.multi_file = True
                    self.first_file_name = self.filename = multi_file.format(0)
                    self.multi_filename = multi_file
                    self.no_subdir = False
                elif os.path.exists(multi_wo_dir.format(0)):
                    self.multi_file = True
                    self.first_file_name = self.filename = multi_wo_dir.format(0)
                    self.multi_filename = multi_file
                    self.no_subdir = True
                else:
                    err_msg = "File not found. Tried locations:\n{}\n{}\n{}"
                    err_msg = err_msg.format(single_file, multi_file.format(0),
                                             multi_wo_dir.format(0))
                    raise FileNotFoundError(err_msg)

        with h5py.File(self.filename, 'r') as f:
            self.Header = dict(f['Header'].attrs)
            self.Config = dict(f['Config'].attrs)
            self.Parameters = dict(f['Parameters'].attrs)
            keys = list(f.keys())

        # Enable units
            self.get_units_and_other_parameters()
        if settings.use_units:
            self.enable_units()
            self.add_user_units()

        # Find the adiabatic index
        self.gamma = self.get_adiabiatic_index()

        # Load all data sets
        if load_all:
            for key in keys:
                self.load_data(key)

        self.load_org_info()

    def load_org_info(self):
        """
        Load some extra info about original data.

        :meta private:
        """
        with h5py.File(self.filename, 'r') as f:
            if 'org_info' in f:
                self['org_info'] = {}
                for key in f['org_info'].attrs.keys():
                    self['org_info'][key] = f['org_info'].attrs[key]

    def get_units_and_other_parameters(self):
        """
        Define arepo units, scale factor (a) and h (HubbleParam).

        The input required is a hdf5 file with the Parameters and Header
        groups found in arepo snapshots.

        :meta private:
        """
        self._Time = self.Header['Time']
        self._HubbleParam = self.Parameters['HubbleParam']

        ComovingIntegrationOn = self.Parameters['ComovingIntegrationOn']

        self.ComovingIntegrationOn = bool(ComovingIntegrationOn)
        self.comoving_sim = self.ComovingIntegrationOn

        if self.ComovingIntegrationOn:
            self._Redshift = self.Header['Redshift']

            Omega0 = self.Parameters['Omega0']
            OmegaBaryon = self.Parameters['OmegaBaryon']
            OmegaLambda = self.Parameters['OmegaLambda']

            # Set up LambdaCDM cosmology to calculate times, etc
            self.cosmo = LambdaCDM(H0=100 * self.h, Om0=Omega0,
                                   Ob0=OmegaBaryon, Ode0=OmegaLambda)
            # Current age of the universe and look back time
            self._age = self.get_age(self.z)
            self._lookback_time = self.get_lookback_time(self.z)

        missing_unitinfo = True
        if 'UnitLength_in_cm' in self.Parameters:
            if 'UnitMass_in_g' in self.Parameters:
                if 'UnitVelocity_in_cm_per_s' in self.Parameters:
                    missing_unitinfo = False

        if settings.use_units and missing_unitinfo:
            raise RuntimeError("Units mssing in self.Parameters!")

        if not missing_unitinfo:
            unit_length = self.Parameters['UnitLength_in_cm'] * u.cm
            unit_mass = self.Parameters['UnitMass_in_g'] * u.g
            unit_velocity = self.Parameters['UnitVelocity_in_cm_per_s'] * u.cm / u.s
            unit_time = unit_length / unit_velocity
            unit_energy = unit_mass * unit_velocity**2
            unit_pressure = (unit_mass / unit_length) / unit_time**2
            unit_density = unit_mass / unit_length**3

            self.arepo_units_in_cgs = {'unit_length': unit_length,
                                       'unit_mass': unit_mass,
                                       'unit_velocity': unit_velocity,
                                       'unit_time': unit_time,
                                       'unit_energy': unit_energy,
                                       'unit_pressure': unit_pressure,
                                       'unit_density': unit_density}

        if settings.use_units:

            _ns = globals()

            try:
                arepo_mass = u.def_unit(
                    ["arepo_mass"],
                    unit_mass,
                    prefixes=False,
                    namespace=_ns,
                    doc="Arepo mass unit",
                    format={"latex": r"arepo\_mass"},
                )
                arepo_time = u.def_unit(
                    ["arepo_time"],
                    unit_time,
                    prefixes=False,
                    namespace=_ns,
                    doc="Arepo time unit",
                    format={"latex": r"arepo\_time"},
                )
                arepo_length = u.def_unit(
                    ["arepo_length"],
                    unit_length,
                    prefixes=False,
                    namespace=_ns,
                    doc="Arepo length unit",
                    format={"latex": r"arepo\_length"},
                )
                arepo_velocity = u.def_unit(
                    ["arepo_velocity"],
                    unit_velocity,
                    prefixes=False,
                    namespace=_ns,
                    doc="Arepo velocity unit",
                    format={"latex": r"arepo\_velocity"},
                )
                arepo_pressure = u.def_unit(
                    ["arepo_pressure"],
                    unit_pressure,
                    prefixes=False,
                    namespace=_ns,
                    doc="Arepo pressure unit",
                    format={"latex": r"arepo\_pressure"},
                )
                arepo_energy = u.def_unit(
                    ["arepo_energy"],
                    unit_energy,
                    prefixes=False,
                    namespace=_ns,
                    doc="Arepo energy unit",
                    format={"latex": r"arepo\_energy"},
                )
                arepo_density = u.def_unit(
                    ["arepo_density"],
                    unit_density,
                    prefixes=False,
                    namespace=_ns,
                    doc="Arepo density unit",
                    format={"latex": r"arepo\_density"},
                )
            except ValueError:
                traceback.print_exc()
                err_msg = ("Re-declaration of arepo code units attempted "
                           + "within same Python session. "
                           + "See https://github.com/tberlok/paicos/issues/59")
                raise RuntimeError(err_msg)

            self.arepo_units = {'unit_length': arepo_length,
                                'unit_mass': arepo_mass,
                                'unit_velocity': arepo_velocity,
                                'unit_time': arepo_time,
                                'unit_energy': arepo_energy,
                                'unit_pressure': arepo_pressure,
                                'unit_density': arepo_density}

    def enable_units(self):
        """
        Enables arepo units globally.

        :meta private:
        """
        for unit_name, unit in self.arepo_units.items():
            u.add_enabled_units(unit)
            phys_type = unit_name.split('_')[1]
            u.def_physical_type(unit, phys_type)

        self._length = self.get_paicos_quantity(1, 'Coordinates')
        self._mass = self.get_paicos_quantity(1, 'Masses')
        self._velocity = self.get_paicos_quantity(1, 'Velocities')

    @property
    def length(self):
        """The unit of length used in the simulation."""
        return self._length

    @property
    def mass(self):
        """The unit of mass used in the simulation."""
        return self._mass

    @property
    def velocity(self):
        """
        One of the units used for velocities in the simulation.

        Note: The a and h-scalings are not the same for velocities
        in the halo catalogs and in the snapshots.
        """
        return self._velocity

    def add_user_units(self):
        """
        Add all user supplied units

        :meta private:
        """
        # pylint: disable=import-outside-toplevel, consider-using-dict-items
        from .. import unit_specifications

        unit_dict = unit_specifications.unit_dict
        user_unit_dict = util.user_unit_dict

        err_msg = ("\n\nThe user supplied unit for '{}:{}' already exists "
                   + "in the default Paicos settings. Changing from '{}' to  "
                   + "'{}' is not allowed. Please make a pull request if you "
                   + " have found a bug.")

        for field in user_unit_dict:
            for name in user_unit_dict[field]:

                # Check for overwriting and changing of defaults
                if name in unit_dict[field]:
                    def_unit = unit_dict[field][name]
                    user_unit = user_unit_dict[field][name]
                    if user_unit != def_unit:
                        msg = err_msg.format(field, name, def_unit, user_unit)
                        del user_unit_dict[field][name]
                        raise RuntimeError(msg)

                # Set the new units
                unit_dict[field][name] = user_unit_dict[field][name]

    def get_adiabiatic_index(self):
        """
        Returns the adiabatic index
        """
        if 'GAMMA' in self.Config:
            gamma = self.Config['GAMMA']
        elif 'ISOTHERMAL' in self.Config:
            gamma = 1
        else:
            gamma = 5 / 3
        return gamma

    @property
    def a(self):
        """
        The scale factor.
        """
        if self.comoving_sim:
            return self._Time
        raise RuntimeError('Non-comoving object has no scale factor')

    @property
    def h(self):
        """
        The reduced Hubble parameter
        """
        if self._HubbleParam == 0:
            return 1.0
        return self._HubbleParam

    @property
    def z(self):
        """
        The redshift.
        """
        if self.comoving_sim:
            return self._Redshift
        raise RuntimeError('Non-comoving object has no redshift')

    @property
    def lookback_time(self):
        """ The lookback time. """
        if self.comoving_sim:
            return self._lookback_time
        raise RuntimeError('Non-comoving object has no lookback_time')

    @property
    def age(self):
        """ The age of the universe in the simulation. """
        if self.comoving_sim:
            return self._age
        raise RuntimeError('Non-comoving object has no lookback_time')

    @property
    def time(self):
        """
        The time elapsed since the beginning of the simulation.

        Only defined for non-comoving simulations.
        """
        if self.comoving_sim:
            msg = 'time not defined for comoving sim'
            raise RuntimeError(msg)
        return self._Time * self.arepo_units['unit_time']

    def rho_crit(self, z):
        """
        Returns the physical critical density (no a or h factors)
        """
        rho_crit = self.cosmo.critical_density(z).to('arepo_density')
        return self.__convert_to_paicos(rho_crit, z)

    def get_lookback_time(self, z):
        """
        Returns the lookback time for a given redshift, z.
        """
        lookback_time = self.cosmo.lookback_time(z)
        return self.__convert_to_paicos(lookback_time, z)

    def get_age(self, z):
        """
        Returns the age of the universe for a given redshift, z.
        """
        age = self.cosmo.lookback_time(1e100) - self.cosmo.lookback_time(z)

        return self.__convert_to_paicos(age, z)

    def __convert_to_paicos(self, time, z):
        """
        Helper function to convert astropy quantities in to PaicosQuantities.

        Really just to still have access to the .label() method.
        """
        if settings.use_units:
            a = 1.0 / (z + 1.)
            if isinstance(a, np.ndarray):
                time = pu.PaicosTimeSeries(time, a=a, h=self.h,
                                           comoving_sim=self.comoving_sim)
            else:
                time = pu.PaicosQuantity(time, a=a, h=self.h,
                                         comoving_sim=self.comoving_sim)
        return time

    def get_paicos_quantity(self, data, name, field='default'):
        """
        Convert some data to a PaicosQuantity.

        Parameters:

            data: typically some numpy array, integer or float

            name: corresponds to a block name in Arepo snapshots,
                  i.e., a key in one of the dictionaries defined
                  in unit_specifications.py

            field: the name of of one of the dictionaries defined in the
                   unit specifications, e.g,:
                    ['default', 'voronoi_cells', 'dark_matter',
                    'stars', 'black_holes', 'groups', 'subhalos']

        Returns: A PaicosQuantity
        """

        if hasattr(data, 'unit'):
            msg = f'Data already had units! {name}'
            raise RuntimeError(msg)

        unit = self.find_unit(name, field)

        if not isinstance(data, np.ndarray):
            data = np.array(data)

        if unit:
            return pu.PaicosQuantity(data, unit, a=self._Time, h=self.h,
                                     comoving_sim=self.comoving_sim,
                                     dtype=data.dtype)
        return data

    def find_unit(self, name, field):
        """
        Find unit for a given quantity.

        Parameters:

            name: corresponds to a block name in Arepo snapshots,
                  i.e., a key in one of the dictionaries defined
                  in unit_specifications.py

            field: the name of of one of the dictionaries defined in the
                   unit specifications, e.g,:
                    ['default', 'voronoi_cells', 'dark_matter',
                    'stars', 'black_holes', 'groups', 'subhalos']
        """
        # pylint: disable=import-outside-toplevel

        # This import statement can only be done after arepo units have
        # been globally enabled
        from .. import unit_specifications

        if field not in unit_specifications.unit_dict:
            raise RuntimeError(f'unknown field: {field}')

        if name not in unit_specifications.unit_dict[field]:
            unit = False
        else:
            unit = unit_specifications.unit_dict[field][name]

        # Convert string to astropy unit
        if isinstance(unit, str):
            unit = u.Unit(unit)

        if unit is False:
            msg = ('\n\nUnit for field:{}, blockname:{} not implemented!'
                   + '\nPlease add it to unit_specifications '
                   + 'by using the pa.add_user_units '
                   + 'function. You can also add the pa.add_user_units call to your'
                   + 'Paicos user settings to avoid having to do this more than once. '
                   + 'Alternatively, please create an issue on GitHub if you think '
                   + 'others could benefit from that.'
                   + '\nFurther instructions can be '
                   + 'found by runnning pa.add_user_units? '
                   + 'in a terminal or notebook.')

            if settings.strict_units:
                raise RuntimeError(msg.format(field, name))
            return False

        return self._sanitize_unit(unit)

    def unit_quantity(self, astropy_unit_str):
        """
        Returns a Paicos quantity with value 1 and any
        astropy unit.
        """
        unit = u.Unit(astropy_unit_str)
        return pu.PaicosQuantity(1.0, unit, a=self._Time, h=self.h,
                                 comoving_sim=self.comoving_sim,
                                 dtype=float)

    def uq(self, astropy_unit_str):
        """
        A short hand for the unit_quantity method.
        """
        return self.unit_quantity(astropy_unit_str)

    def _sanitize_unit(self, unit):
        """
        Removes 'a' factors for non-comoving simulations,
        and 'h' factors for simulations with HubbleParam=1.
        """

        remove_list = []
        if not self.ComovingIntegrationOn:
            remove_list.append(u.Unit('small_a'))

        if self.h == 1.:
            remove_list.append(u.Unit('small_h'))

        return pu.get_new_unit(unit, remove_list)

    def load_data(self, name, group=None):
        """
        Load data from a generic Paicos hdf5 file (written by a PaicosWriter
        instance).

        The method requires that the data sets have a 'unit' attribute
        and hence does not work for Arepo hdf5 files.
        For this reason, this method is overloaded in the Snapshot and Catalog
        classes.
        """

        with h5py.File(self.filename, 'r') as f:
            if isinstance(f[name], h5py.Dataset):
                self[name] = util.load_dataset(f, name, group=group)
                if isinstance(self[name], pu.PaicosQuantity) and self.to_physical:
                    self[name] = self[name].to_physical
            elif isinstance(f[name], h5py.Group):
                for data_name in f[name].keys():
                    data = util.load_dataset(f, data_name, group=name)
                    if name not in self:
                        self[name] = {}
                    if isinstance(data, pu.PaicosQuantity) and self.to_physical:
                        self[name][data_name] = data.to_physical
                    else:
                        self[name][data_name] = data


class ImageReader(PaicosReader):
    """
    This is a subclass of the PaicosReader.

    It reads the additional information stored in image files and makes
    them accessible as attributes, i.e., extent, widths, center, direction
    and image_creator.

    For projection files, it also tries to automatically get derived
    variables, e.g., if the hdf5 file contains

    'MagneticFieldSquaredTimesVolume' and 'Volume'

    then it will automatically divide them to obtain the MagneticFieldSquared.
    """

    def __init__(self, basedir='.', snapnum=None, basename="projection", load_all=True):
        """
        See documentation for the PaicosReader.

        Returns a dictionary with additional attributes.
        """

        # The PaicosReader class takes care of most of the loading
        super().__init__(basedir, snapnum, basename=basename,
                         load_all=load_all)

        # Load info specific to images
        with h5py.File(self.filename, 'r') as f:
            self.extent = util.load_dataset(f, 'extent', group='image_info')
            self.widths = util.load_dataset(f, 'widths', group='image_info')
            self.center = util.load_dataset(f, 'center', group='image_info')
            self.direction = direction = f['image_info'].attrs['direction']
            self.image_creator = f['image_info'].attrs['image_creator']

            # Recreate orientation object if it is saved
            if direction == 'orientation':
                normal_vector = f['image_info'].attrs['normal_vector']
                perp_vector1 = f['image_info'].attrs['perp_vector1']
                self.orientation = Orientation(normal_vector=normal_vector,
                                               perp_vector1=perp_vector1)
            else:
                self.orientation = None

        if self.direction == 'orientation':
            self._im = ImageCreator(self, self.center, self.widths, self.orientation)
        else:
            self._im = ImageCreator(self, self.center, self.widths, self.direction)

        self.x_c = self.center[0]
        self.y_c = self.center[1]
        self.z_c = self.center[2]
        self.width_x = self.widths[0]
        self.width_y = self.widths[1]
        self.width_z = self.widths[2]

        self.direction = direction

        self.width = self._im.width
        self.height = self._im.height
        self.depth = self._im.depth
        self.centered_extent = self._im.centered_extent
        self.area = self._im.area
        self.volume = self._im.volume

        if len(list(self.keys())) > 0:
            arr_shape = self[list(self.keys())[0]].shape
            self.npix = self.npix_width = arr_shape[0]
            self.npix_height = arr_shape[1]

        self.area_per_pixel = self.area / (self.npix_width * self.npix_height)
        self.volume_per_pixel = self.volume / (self.npix_width * self.npix_height)

        self.dw = self.width / self.npix_width
        self.dh = self.height / self.npix_height

        # Get derived images
        self.get_derived_images()

    @np.errstate(divide='ignore', invalid='ignore')
    def get_derived_images(self):
        """
        Calculate images automatically for convenience,
        e.g., 0_MagneticFieldSquaredTimesVolume and 0_Volume
        are divided to obtain a 2D-array with 0_MagneticFieldSquared
        """

        # Get derived images from projection-files
        keys = list(self.keys())
        for key in keys:
            if 'Times' in key:
                # Keys of the form 'MagneticFieldSquaredTimesVolume'
                # are split up
                start, end = key.split('Times')
                if end in keys:
                    self[start] = self[key] / self[end]
                elif start[0:2] + end in keys:
                    self[start] = self[key] / self[start[0:2] + end]

        # Calculate density if we have both masses and volumes
        for p in ['', '0_']:
            if (p + 'Masses' in keys) and (p + 'Volume' in keys):
                self[p + 'Density'] = self[p + 'Masses'] / self[p + 'Volume']

    def get_image_coordinates(self):
        extent = self.extent
        npix_width = self.npix_width
        npix_height = self.npix_height
        width = self.width
        height = self.height

        w = extent[0] + (np.arange(npix_width) + 0.5) * width / npix_width
        h = extent[2] + (np.arange(npix_height) + 0.5) * height / npix_height

        if settings.use_units:
            wu = w.unit_quantity
            ww, hh = np.meshgrid(w.value, h.value)
            ww = ww * wu
            hh = hh * wu
        else:
            ww, hh = np.meshgrid(w, h)

        return ww, hh

    def get_centered_image_coordinates(self):

        extent = self.centered_extent
        npix_width = self.npix_width
        npix_height = self.npix_height
        width = self.width
        height = self.height

        w = extent[0] + (np.arange(npix_width) + 0.5) * width / npix_width
        h = extent[2] + (np.arange(npix_height) + 0.5) * height / npix_height

        if settings.use_units:
            wu = w.unit_quantity
            ww, hh = np.meshgrid(w.value, h.value)
            ww = ww * wu
            hh = hh * wu
        else:
            ww, hh = np.meshgrid(w, h)

        return ww, hh


class Histogram2DReader(PaicosReader):
    """
    This is a subclass of the PaicosReader.

    It reads the additional information stored by a Histogram2D instance
    and makes them accessible as attributes, i.e., colorlabel, normalize,
    logscale, hist2d, centers_x, centers_y.
    """

    def __init__(self, basedir='.', snapnum=None, basename='2d_histogram'):
        """
        See documentation for the PaicosReader.

        Returns a dictionary with additional attributes.
        """

        # The PaicosReader class takes care of most of the loading
        super().__init__(basedir, snapnum, basename=basename,
                         load_all=True)

        with h5py.File(self.filename, 'r') as hdf5file:
            if 'colorlabel' in hdf5file['hist2d'].attrs.keys():
                self.colorlabel = hdf5file['hist2d'].attrs['colorlabel']
            self.normalize = hdf5file['hist_info'].attrs['normalize']
            self.logscale = hdf5file['hist_info'].attrs['logscale']

        self.hist2d = self['hist2d']
        self.centers_x = self['centers_x']
        self.centers_y = self['centers_y']



import numpy as np
import astropy.units as u
import traceback
import numbers
from .paicos_readers import PaicosReader
from .. import settings
from .. import units as pu


class GenericSnapshot(PaicosReader):
    """
    This is a very simple Snapshot class, which
    was written for use with general data (e.g. if one
    wants to make images of Ramses simulations).

    Detailed documentation/example usage remains to be written,
    so please create an issue if you have trouble getting this to work.
    """

    def __init__(self, basedir='.', snapnum=None, basename="snap",
                 basesubdir='snapdir', load_all=True, to_physical=False,
                 only_init=False,
                 verbose=False):
        """
        Please use see the docstring for the method 'give_info'.
        """

        try:
            super().__init__(basedir, snapnum, basename, basesubdir,
                             load_all, to_physical, verbose)
            self.box = self.Header["BoxSize"]
            self.box_size = np.array([self.box, self.box, self.box])
            if settings.use_units:
                self.box_size = self.box_size * self.length
        except FileNotFoundError as e:
            # File not found, user will have to call give_info
            if not only_init:
                import warnings
                msg = ('\n\nset only_init=True to initialize without'
                       + ' loading a data file. You will then need'
                       + ' to call give_info instead\n\n')
                warnings.warn(msg)
                raise e

        self._auto_list = []

    def give_info(self, boxsize_in_code_units, time_in_code_units,
                  snapnum=None,
                  length_unit=None,
                  time_unit=None,
                  mass_unit=None,
                  comoving_sim=False,
                  hubble_param=1,
                  redshift=0,
                  Omega0=1,
                  OmegaLambda=0,
                  OmegaBaryon=1
                  ):

        """
        Parameters
        ----------

            snapnum : int
                The snapshot number that you are analyzing

            boxsize_in_code_units : float
                The box size of your simulation box, which is assumed to be cubic.


        Optional parameters (needed when using units are enabled)
        --------------------------

            time_in_code_units : float

            length_unit : astropy unit
                The code unit used in the simulation.
                For length e.g. 0.2 * u.Unit('kpc')

            time_unit : astropy unit
                The code unit used in the simulation.
                For time e.g. u.Unit('Myr')

            mass_unit : astropy unit
                The code unit used in the simulation.
                For mass e.g. u.Unit('Msun')
        """
        if hasattr(self, "Parameters"):
            err_msg = ("give_info should not be called"
                       + " if Parameters has already been"
                       + " via a previous call to give_info"
                       + " or by the PaicosReader base class")
            raise RuntimeError(err_msg)

        self.Parameters = {}
        self.Header = {}
        self.Config = {}

        self.snapnum = snapnum

        self.Header['Time'] = time_in_code_units
        self.Parameters['HubbleParam'] = hubble_param
        self.Header["BoxSize"] = boxsize_in_code_units

        if settings.use_units:
            err_msg = ('It is expected that you pass'
                       + ' this unit when pa.settings.use_units is True')

            assert length_unit is not None, err_msg
            assert time_unit is not None, err_msg
            assert mass_unit is not None, err_msg

            unit_length = 1.0 * length_unit
            unit_time = 1.0 * time_unit
            unit_mass = 1.0 * mass_unit
            assert isinstance(unit_length, u.Quantity)
            assert isinstance(unit_time, u.Quantity)
            assert isinstance(unit_mass, u.Quantity)

            unit_velocity = unit_length / unit_time

            self.Parameters['UnitLength_in_cm'] = unit_length.to('cm').value
            self.Parameters['UnitMass_in_g'] = unit_mass.to('g').value
            self.Parameters['UnitVelocity_in_cm_per_s'] = unit_velocity.to('cm/s').value

        if comoving_sim:
            self.Header['Redshift'] = redshift
            self.Parameters['ComovingIntegrationOn'] = 1
            self.Parameters['Omega0'] = Omega0
            self.Parameters['OmegaBaryon'] = OmegaBaryon
            self.Parameters['OmegaLambda'] = OmegaLambda
        else:
            self.Parameters['ComovingIntegrationOn'] = 0

        # Enable units
        self.get_units_and_other_parameters()
        if settings.use_units:
            self.enable_units()

        self.box = self.Header["BoxSize"]
        self.box_size = np.array([self.box, self.box, self.box])
        if settings.use_units:
            self.box_size = self.box_size * self.length

        if self.comoving_sim:
            raise RuntimeError("comoving_sim not tested - please create an issue!")

        if self._HubbleParam != 1:
            raise RuntimeError("h!=1 not tested - please create an issue!")

    def set_volumes(self, data, unit='arepo_length^3'):
        """
        Use this to set the cell volumes.
        """
        self.set_data(data, '0_Volume', unit)

    def set_positions(self, data, unit='arepo_length'):
        """
        Use this to set the cell positions.
        """
        assert data.shape[1] == 3
        self.set_data(data, '0_Coordinates', unit)

    def set_data(self, data, key, unit=None):
        """
        Set data in the snapshot

        Parameters
        ----------

            data : numpy array
                For instance the density, temperature etc

            key : str
                The dictionary key.

            unit : astropy unit
                You only need to set this if you have units turned on.
        """
        import warnings
        if settings.double_precision:
            # Load all variables with double precision
            if not issubclass(data.dtype.type, numbers.Integral):
                data = data.astype(np.float64)
        else:
            warnings.warn('\n\nThe cython routines expect double precision '
                          + 'and will fail unless settings.double_precision '
                          + 'is True.\n\n')
        if settings.use_units:
            assert unit is not None
            self[key] = pu.PaicosQuantity(data, unit, h=self._HubbleParam, a=self._Time,
                                          comoving_sim=self.comoving_sim)
        else:
            self[key] = data



import numpy as np
import cupy as cp
from numba import cuda
import numba

from .image_creator import ImageCreator
from .. import util
from .. import settings
from .. import units

from ..trees.bvh_gpu import GpuBinaryTree
from ..trees.bvh_gpu import nearest_neighbor_device
from ..trees.bvh_gpu import is_point_in_box, distance


def get_blocks(size, threadsperblock):
    return (size + (threadsperblock - 1)) // threadsperblock


@cuda.jit(device=True, inline=True)
def rotate_point_around_center(point, tmp_point, center, rotation_matrix):
    """
    Rotate around center. Note that we overwrite point.
    """

    # Subtract center
    for ii in range(3):
        tmp_point[ii] = point[ii] - center[ii]
        point[ii] = 0.0

    # Rotate around center (matrix multiplication).
    for ii in range(3):
        for jj in range(3):
            point[ii] += rotation_matrix[ii, jj] * tmp_point[jj]

    # Add center back
    for ii in range(3):
        point[ii] = point[ii] + center[ii]


@cuda.jit
def trace_rays(points, tree_parents, tree_children, tree_bounds, variable, hsml,
               widths, center,
               tree_scale_factor, tree_offsets, image, rotation_matrix, tol):

    ix, iy = cuda.grid(2)

    nx = image.shape[0]
    ny = image.shape[1]

    if ix >= 0 and ix < nx and iy >= 0 and iy < ny:
        result = 0.0

        dx = widths[0] / nx
        dy = widths[1] / ny

        num_internal_nodes = tree_children.shape[0]

        # Initialize z and dz (in arepo code units)
        z = 0.0

        query_point = numba.cuda.local.array(3, numba.float64)
        tmp_point = numba.cuda.local.array(3, numba.float64)

        while z < widths[2]:

            # Query points in aligned coords
            query_point[0] = (center[0] - widths[0] / 2.0) + (ix + 0.5) * dx
            query_point[1] = (center[1] - widths[1] / 2.0) + (iy + 0.5) * dy
            query_point[2] = (center[2] - widths[2] / 2.0) + z

            # Rotate to simulation coords
            rotate_point_around_center(
                query_point, tmp_point, center, rotation_matrix)

            # Convert to the tree coordinates
            query_point[2] = (query_point[2] - tree_offsets[2]
                              ) * tree_scale_factor
            query_point[0] = (query_point[0] - tree_offsets[0]
                              ) * tree_scale_factor
            query_point[1] = (query_point[1] - tree_offsets[1]
                              ) * tree_scale_factor

            min_dist, min_index = nearest_neighbor_device(points, tree_parents, tree_children,
                                                          tree_bounds, query_point,
                                                          num_internal_nodes)

            # Calculate dz
            dz = tol * hsml[min_index]

            # Update integral
            result = result + dz * variable[min_index]

            # Update position
            z = z + dz

        # Subtract the 'extra' stuff added in last iteration
        result = result - (z - widths[2]) * variable[min_index]
        # Set result in image array
        image[ix, iy] = result


class GpuRayProjector(ImageCreator):
    """
    A class that allows creating an image of a given variable by projecting
    it onto a 2D plane. This class works by raytracing the variable
    (i.e. by calculating a line integral along the line-of-sight).

    It only works on cuda-enabled GPUs.
    """

    def __init__(self, snap, center, widths, direction,
                 npix=512, parttype=0, tol=0.25, threadsperblock=8,
                 do_pre_selection=False):
        """
        Initialize the Projector class.

        Parameters
        ----------
        snap : Snapshot
            A snapshot object of Snapshot class from paicos package.

        center : numpy array
            Center of the region on which projection is to be done, e.g.
            center = [x_c, y_c, z_c].

        widths : numpy array
            Widths of the region on which projection is to be done,
            e.g.m widths=[width_x, width_y, width_z].

        direction : str
            Direction of the projection, e.g. 'x', 'y' or 'z',
            or a Paicos Orientation class instance.

        npix : int, optional
            Number of pixels in the horizontal direction of the image,
            by default 512.

        parttype : int, optional
            Number of the particle type to project, by default gas (PartType 0).

        """

        # call the superclass constructor to initialize the ImageCreator class
        super().__init__(snap, center, widths, direction, npix=npix, parttype=parttype)

        parttype = self.parttype

        self.threadsperblock = threadsperblock

        self.do_pre_selection = do_pre_selection

        self.tol = tol

        # Calculate the smoothing length
        avail_list = (list(snap.keys()) + snap._auto_list)
        if f'{parttype}_Volume' in avail_list:
            self.hsml = 2.0 * np.cbrt((self.snap[f"{parttype}_Volume"])
                                      / (4.0 * np.pi / 3.0))
        elif f'{parttype}_SubfindHsml' in avail_list:
            self.hsml = self.snap[f'{parttype}_SubfindHsml']
        else:
            raise RuntimeError(
                'There is no smoothing length or volume for the projector')

        self.pos = self.snap[f'{self.parttype}_Coordinates']

        if settings.use_units:
            self.hsml = self.hsml.to(self.pos.unit)

        # Call selection
        self.has_do_region_selection_been_called = False
        self._do_region_selection()
        self.has_do_region_selection_been_called = True

    def _do_region_selection(self):

        self.do_unit_consistency_check()

        center = self.center
        widths = self.widths
        snap = self.snap
        parttype = self.parttype

        if self.has_do_region_selection_been_called:
            if self.do_pre_selection:
                import warnings
                err_msg = ("It looks like you are changing projector "
                           + " properties after the fact, i.e. changing widths "
                           + "center, orientation, resolution etc. This might be "
                           + "slow with with the option do_pre_selection, which "
                           + "you have turned on. If your GPU has enough memory "
                           + "then it is probably better to set do_pre_selection "
                           + "to False.")
                warnings.warn(err_msg)

        # Send subset of snapshot to GPU
        if self.do_pre_selection:
            # get the index of the region of projection
            if self.direction != 'orientation':
                get_index = util.get_index_of_cubic_region_plus_thin_layer
                self.index = get_index(self.snap[f"{parttype}_Coordinates"],
                                       center, widths, 2.0 * self.hsml, snap.box)
            else:
                get_index = util.get_index_of_rotated_cubic_region_plus_thin_layer
                self.index = get_index(snap[f"{parttype}_Coordinates"],
                                       center, widths, 2.0 * self.hsml, snap.box,
                                       self.orientation)

            self.hsml = self.hsml[self.index]
            self.pos = self.pos[self.index]

            self._send_data_to_gpu()

            # We need to reconstruct the tree!
            self.tree = GpuBinaryTree(self.gpu_variables['pos'],
                                      self.gpu_variables['hsml'])
            del self.gpu_variables['pos']
            self.gpu_variables['hsml'] = self.gpu_variables['hsml'][self.tree.sort_index]
        # Send entirety of snapshot to GPU (if we have not already
        # done so). Always send small data with change in resolution etc
        else:
            if not self.has_do_region_selection_been_called:
                self._send_data_to_gpu()

                # Construct tree
                self.tree = GpuBinaryTree(self.gpu_variables['pos'],
                                          self.gpu_variables['hsml'])

                del self.gpu_variables['pos']
                self.gpu_variables['hsml'] = self.gpu_variables['hsml'][
                    self.tree.sort_index]

        # Always send small data
        self._send_small_data_to_gpu()

    def _send_data_to_gpu(self):
        self.gpu_variables = {}
        if settings.use_units:
            self.gpu_variables['pos'] = cp.array(self.pos.value)
            self.gpu_variables['hsml'] = cp.array(self.hsml.value)
        else:
            self.gpu_variables['pos'] = cp.array(self.pos)
            self.gpu_variables['hsml'] = cp.array(self.hsml)

        self._send_small_data_to_gpu()

    def _send_small_data_to_gpu(self):

        self.gpu_variables['rotation_matrix'] = cp.array(
            self.orientation.rotation_matrix)

        if settings.use_units:
            self.gpu_variables['widths'] = cp.array(self.widths.value)
            self.gpu_variables['center'] = cp.array(self.center.value)
        else:
            self.gpu_variables['widths'] = cp.array(self.widths)
            self.gpu_variables['center'] = cp.array(self.center)

    def _gpu_project(self, variable_str):
        """
        Private method for projecting using cuda code
        """
        rotation_matrix = self.gpu_variables['rotation_matrix']
        widths = self.gpu_variables['widths']
        center = self.gpu_variables['center']
        hsml = self.gpu_variables['hsml']
        nx = self.npix_width
        ny = self.npix_height
        variable = self.gpu_variables[variable_str]

        tree_scale_factor = self.tree.conversion_factor
        tree_offsets = self.tree.off_sets

        image = cp.zeros((nx, ny))

        blocks_x = get_blocks(nx, self.threadsperblock)
        blocks_y = get_blocks(ny, self.threadsperblock)
        btuple = (blocks_x, blocks_y)
        ttuple = (self.threadsperblock, self.threadsperblock)
        trace_rays[btuple, ttuple](self.tree._pos, self.tree.parents, self.tree.children,
                                   self.tree.bounds, variable, hsml, widths, center,
                                   tree_scale_factor, tree_offsets, image,
                                   rotation_matrix, self.tol)

        projection = cp.asnumpy(image)
        return projection

    def project_variable(self, variable, additive=False):
        """
        projects a given variable onto a 2D plane.

        Parameters
        ----------
        variable : str, function, numpy array
            variable, it can be passed as string or an array

        Returns
        -------
        numpy array
            The image of the projected variable
        """

        # This calls _do_region_selection if resolution, Orientation,
        # widths or center changed
        self._check_if_properties_changed()

        if additive:
            err_msg = "GPU ray tracer does not yet support additive=True"
            raise RuntimeError(err_msg)

        if isinstance(variable, str):
            variable_str = str(variable)
            err_msg = 'projector uses a different parttype'
            assert int(variable[0]) == self.parttype, err_msg
            variable = self.snap[variable]
        else:
            variable_str = 'projection_variable'
            if not isinstance(variable, np.ndarray):
                raise RuntimeError('Unexpected type for variable')

        assert len(variable.shape) == 1, 'only scalars can be projected'

        # Select same part of array that the projector has selected
        if self.do_pre_selection:
            variable = variable[self.index]

        if variable_str in self.gpu_variables and variable_str != 'projection_variable':
            pass
        else:
            # Send variable to gpu
            if settings.use_units:
                self.gpu_variables[variable_str] = cp.array(variable.value)
            else:
                self.gpu_variables[variable_str] = cp.array(variable)

            # Sort the variable according to Morton code sorting
            self.gpu_variables[variable_str] = self.gpu_variables[variable_str][
                self.tree.sort_index]

        # Do the projection
        projection = self._gpu_project(variable_str)

        # Transpose
        projection = projection.T

        assert projection.shape[0] == self.npix_height
        assert projection.shape[1] == self.npix_width

        if isinstance(variable, units.PaicosQuantity):
            unit_length = self.snap['0_Coordinates'].uq
            projection = projection * variable.unit_quantity * unit_length

        return projection / self.depth

    def __del__(self):
        """
        Clean up like this? Not sure it is needed...
        """
        del self.gpu_variables
        del self.tree
        cp._default_memory_pool.free_all_blocks()



from cython.parallel import prange, parallel
cimport openmp
import numpy as np
cimport numpy as np

ctypedef fused real_t:
    float
    double
    np.float32_t
    np.float64_t

STUFF = "Hi"  # https://stackoverflow.com/questions/8024805/cython-compiled-c-extension-importerror-dynamic-module-does-not-define-init-fu


def project_image(real_t[:] xvec, real_t[:] yvec, real_t[:] variable,
                  real_t[:] hvec, int nx, real_t xc, real_t yc,
                  real_t sidelength_x, real_t sidelength_y,
                  real_t boxsize, int numthreads=1):

    """

    Projects particles using an SPH-like kernel.

    This function assumes that only particles with coordinates

    x0 < x < x0 + sidelength_x
    y0 < y < y0 + sidelength_y

    are passed to it. For speed, this is not checked!

    Here x0, y0 are the coordinates at the lower, left corner.

    Parameters:
        xvec (array, N): positions along x (horizontal)
        yvec (array, N): positions along y (vertical)
        variable (array, N): variable to be projected (e.g. mass)
        hsml (array, N): size of particles
        sidelength_x (double): size of image along x
        sidelength_y (double): size of image along y
        boxsize (double): size of simulation domain,
                           which for now is assumed to be cubic!

    Returns:
        2d array: A 2D array with the projected variable.

    """

    assert numthreads == 1, 'use project_image_omp for more than one thread'

    # Number of particles
    cdef int Np = xvec.shape[0]

    # Shape of projection array
    cdef int ny = <int> (sidelength_y/sidelength_x * nx)
    msg = '(sidelength_y/sidelength_x * nx) needs to be an integer'
    assert (sidelength_y/sidelength_x * nx) == <float> ny, msg

    # Create projection array
    cdef real_t[:, :] projection = np.zeros((nx, ny), dtype=np.float64)

    # Loop integers and other variables
    cdef int ip, ix, iy, ih
    cdef int ipx, ipy
    cdef int ix_min, ix_max
    cdef int iy_min, iy_max
    cdef real_t dx, dy, r2, h, h2, weight
    cdef real_t x, y, norm
    cdef real_t x0, y0

    assert sidelength_x/nx == sidelength_y/ny

    # Lower left corner of image in Arepo coordinates
    x0 = xc - sidelength_x / 2.0
    y0 = yc - sidelength_y / 2.0

    for ip in range(Np):
        # Center coordinate system at x0, y0
        x = xvec[ip] - x0
        y = yvec[ip] - y0

        # Apply periodic boundary condition
        # if x < 0.0:
        #     x = x + boxsize
        # elif x > boxsize:
        #     x = x - boxsize

        # if y < 0.0:
        #     y = y + boxsize
        # elif y > boxsize:
        #     y = y - boxsize

        # Position of particle in units of sidelength (0, sidelength)
        x = x*nx/sidelength_x
        y = y*ny/sidelength_y
        h = hvec[ip]*nx/sidelength_x

        if h < 1.:
            h = 1.

        # Index of closest grid point
        ipx = <int> x
        ipy = <int> y

        # Smoothing length as integer
        ih = <int> h + 1

        # Square of smoothing length
        h2 = h*h

        # Find minimum and maximum integers
        ix_min = ipx - ih
        iy_min = ipy - ih
        ix_max = ipx + ih
        iy_max = ipy + ih

        norm = 0.0
        for ix in range(ix_min, ix_max):
            for iy in range(iy_min, iy_max):
                dx = x - 0.5 - <real_t> ix
                dy = y - 0.5 - <real_t> iy
                r2 = dx*dx + dy*dy

                weight = 1.0 - r2/h2
                if weight > 0.0:
                    norm += weight

        # Find minimum and maximum integers
        ix_min = max(0, ipx - ih)
        iy_min = max(0, ipy - ih)
        ix_max = min(projection.shape[0], ipx + ih)
        iy_max = min(projection.shape[1], ipy + ih)

        for ix in range(ix_min, ix_max):
            for iy in range(iy_min, iy_max):
                dx = x - 0.5 - <real_t> ix
                dy = y - 0.5 - <real_t> iy
                r2 = dx*dx + dy*dy

                weight = 1.0 - r2/h2
                if weight > 0.0:
                    projection[ix, iy] += weight*variable[ip]/norm

    # Fix to avoid returning a memory-view
    tmp = np.zeros((nx, ny), dtype=np.float64)
    tmp[:, :] = projection[:, :]
    return tmp


def project_image_omp(real_t[:] xvec, real_t[:] yvec, real_t[:] variable,
                      real_t[:] hvec, int nx, real_t xc, real_t yc,
                      real_t sidelength_x, real_t sidelength_y,
                      real_t boxsize, int numthreads):

    """
    Same as project_image but here with an openmp parallel implementation.
    """

    # Number of particles
    cdef int Np = xvec.shape[0]

    # Shape of projection array
    cdef int ny = <int> (sidelength_y/sidelength_x * nx)
    msg = '(sidelength_y/sidelength_x * nx) needs to be an integer'
    assert (sidelength_y/sidelength_x * nx) == <float> ny, msg

    # Loop integers and other variables
    cdef int ip, ix, iy, ih, threadnum
    cdef int ipx, ipy
    cdef int ix_min, ix_max
    cdef int iy_min, iy_max
    cdef real_t dx, dy, r2, h, h2, weight
    cdef real_t x, y, norm
    cdef real_t x0, y0

    assert sidelength_x/nx == sidelength_y/ny

    # Lower left corner of image in Arepo coordinates
    x0 = xc - sidelength_x/2.0
    y0 = yc - sidelength_y/2.0

    cdef real_t[:, :, :] tmp_var = np.zeros((nx, ny, numthreads), dtype=np.float64)
    cdef real_t[:, :] projection = np.zeros((nx, ny), dtype=np.float64)

    with nogil, parallel(num_threads=numthreads):
        for ip in prange(Np, schedule='static'):
            threadnum = openmp.omp_get_thread_num()
            # Center coordinate system at x0, y0
            x = xvec[ip] - x0
            y = yvec[ip] - y0

            # Apply periodic boundary condition
            # if x < 0.0:
            #     x = x + boxsize
            # elif x > boxsize:
            #     x = x - boxsize

            # if y < 0.0:
            #     y = y + boxsize
            # elif y > boxsize:
            #     y = y - boxsize

            # Position of particle in units of sidelength (0, sidelength)
            x = x*nx/sidelength_x
            y = y*ny/sidelength_y
            h = hvec[ip]*nx/sidelength_x

            if h < 1.:
                h = 1.

            # Index of closest grid point
            ipx = <int> x
            ipy = <int> y

            # Smoothing length as integer
            ih = <int> h + 1

            # Square of smoothing length
            h2 = h*h

            # Find minimum and maximum integers
            ix_min = ipx - ih
            iy_min = ipy - ih
            ix_max = ipx + ih
            iy_max = ipy + ih

            norm = 0.0
            for ix in range(ix_min, ix_max):
                for iy in range(iy_min, iy_max):
                    dx = x - 0.5 - <real_t> ix
                    dy = y - 0.5 - <real_t> iy
                    r2 = dx*dx + dy*dy

                    weight = 1.0 - r2/h2
                    if weight > 0.0:
                        norm = norm + weight

            # Find minimum and maximum integers
            ix_min = max(0, ipx - ih)
            iy_min = max(0, ipy - ih)
            ix_max = min(projection.shape[0], ipx + ih)
            iy_max = min(projection.shape[1], ipy + ih)

            for ix in range(ix_min, ix_max):
                for iy in range(iy_min, iy_max):
                    dx = x - 0.5 - <real_t> ix
                    dy = y - 0.5 - <real_t> iy
                    r2 = dx*dx + dy*dy

                    weight = 1.0 - r2/h2
                    if weight > 0.0:
                        tmp_var[ix, iy, threadnum] = tmp_var[ix, iy, threadnum] \
                                                     + weight*variable[ip]/norm

    # Add up contributions from each thread
    for threadnum in range(numthreads):
        for ix in range(nx):
            for iy in range(ny):
                projection[ix, iy] = projection[ix, iy] + tmp_var[ix, iy, threadnum]

    # Fix to avoid returning a memory-view
    tmp = np.zeros((nx, ny), dtype=np.float64)
    tmp[:, :] = projection[:, :]
    return tmp

def project_oriented_image(real_t[:] xvec, real_t[:] yvec, real_t[:] zvec, real_t[:] variable,
                           real_t[:] hvec, int nx,
                           real_t xc, real_t yc, real_t zc,
                           real_t sidelength_x, real_t sidelength_y,
                           real_t boxsize,
                           real_t[:] unit_vector_x,
                           real_t[:] unit_vector_y,
                           real_t[:] unit_vector_z,
                           int numthreads=1):

    """
    Same as project_image but here for an arbitrarily oriented image.
    """

    assert numthreads == 1, 'use project_image_omp for more than one thread'

    # Number of particles
    cdef int Np = xvec.shape[0]

    # Shape of projection array
    cdef int ny = <int> (sidelength_y/sidelength_x * nx)
    msg = '(sidelength_y/sidelength_x * nx) needs to be an integer'
    assert (sidelength_y/sidelength_x * nx) == <float> ny, msg

    # Create projection array
    cdef real_t[:, :] projection = np.zeros((nx, ny), dtype=np.float64)

    # Loop integers and other variables
    cdef int ip, ix, iy, ih
    cdef int ipx, ipy
    cdef int ix_min, ix_max
    cdef int iy_min, iy_max
    cdef real_t dx, dy, r2, h, h2, weight
    cdef real_t x, y, norm
    cdef real_t cen_x, cen_y, cen_z

    assert sidelength_x/nx == sidelength_y/ny

    for ip in range(Np):
        # Centered coordinates
        cen_x = xvec[ip] - xc
        cen_y = yvec[ip] - yc
        cen_z = zvec[ip] - zc

        # Apply boundary conditions here?

        # Projection of coordinate along perp_vector1
        x = cen_x * unit_vector_x[0] + cen_y * unit_vector_x[1] + cen_z * unit_vector_x[2]
        # Projection of coordinate along perp_vector2
        y = cen_x * unit_vector_y[0] + cen_y * unit_vector_y[1] + cen_z * unit_vector_y[2]

        # Transform these so that they are in the range [0, sidelength]
        x = x + sidelength_x/2.0
        y = y + sidelength_y/2.0

        #----------------------------------------------------------------------------------#
        # Below this line the code is identical to project_image function. Could clean this up.
        #----------------------------------------------------------------------------------#

        # Position of particle in units of sidelength (0, sidelength)
        x = x*nx/sidelength_x
        y = y*ny/sidelength_y
        h = hvec[ip]*nx/sidelength_x

        if h < 1.:
            h = 1.

        # Index of closest grid point
        ipx = <int> x
        ipy = <int> y

        # Smoothing length as integer
        ih = <int> h + 1

        # Square of smoothing length
        h2 = h*h

        # Find minimum and maximum integers
        ix_min = ipx - ih
        iy_min = ipy - ih
        ix_max = ipx + ih
        iy_max = ipy + ih

        norm = 0.0
        for ix in range(ix_min, ix_max):
            for iy in range(iy_min, iy_max):
                dx = x - 0.5 - <real_t> ix
                dy = y - 0.5 - <real_t> iy
                r2 = dx*dx + dy*dy

                weight = 1.0 - r2/h2
                if weight > 0.0:
                    norm += weight

        # Find minimum and maximum integers
        ix_min = max(0, ipx - ih)
        iy_min = max(0, ipy - ih)
        ix_max = min(projection.shape[0], ipx + ih)
        iy_max = min(projection.shape[1], ipy + ih)

        for ix in range(ix_min, ix_max):
            for iy in range(iy_min, iy_max):
                dx = x - 0.5 - <real_t> ix
                dy = y - 0.5 - <real_t> iy
                r2 = dx*dx + dy*dy

                weight = 1.0 - r2/h2
                if weight > 0.0:
                    projection[ix, iy] += weight*variable[ip]/norm

    # Fix to avoid returning a memory-view
    tmp = np.zeros((nx, ny), dtype=np.float64)
    tmp[:, :] = projection[:, :]
    return tmp


def project_oriented_image_omp(real_t[:] xvec, real_t[:] yvec, real_t[:] zvec, real_t[:] variable,
                               real_t[:] hvec, int nx,
                               real_t xc, real_t yc, real_t zc,
                               real_t sidelength_x, real_t sidelength_y,
                               real_t boxsize,
                               real_t[:] unit_vector_x,
                               real_t[:] unit_vector_y,
                               real_t[:] unit_vector_z,
                               int numthreads=1):

    """
    Same as project_oriented_image but here with an openmp implementation.
    """

    # Number of particles
    cdef int Np = xvec.shape[0]

    # Shape of projection array
    cdef int ny = <int> (sidelength_y/sidelength_x * nx)
    msg = '(sidelength_y/sidelength_x * nx) needs to be an integer'
    assert (sidelength_y/sidelength_x * nx) == <float> ny, msg

    # Loop integers and other variables
    cdef int ip, ix, iy, ih, threadnum
    cdef int ipx, ipy
    cdef int ix_min, ix_max
    cdef int iy_min, iy_max
    cdef real_t dx, dy, r2, h, h2, weight
    cdef real_t x, y, norm
    cdef real_t cen_x, cen_y, cen_z

    assert sidelength_x/nx == sidelength_y/ny

    # Create projection arrays
    cdef real_t[:, :, :] tmp_var = np.zeros((nx, ny, numthreads), dtype=np.float64)
    cdef real_t[:, :] projection = np.zeros((nx, ny), dtype=np.float64)

    with nogil, parallel(num_threads=numthreads):
        for ip in prange(Np, schedule='static'):
            threadnum = openmp.omp_get_thread_num()
            # Centered coordinates
            cen_x = xvec[ip] - xc
            cen_y = yvec[ip] - yc
            cen_z = zvec[ip] - zc

            # Apply boundary conditions here?

            # Projection of coordinate along perp_vector1
            x = cen_x * unit_vector_x[0] + cen_y * unit_vector_x[1] + cen_z * unit_vector_x[2]
            # Projection of coordinate along perp_vector2
            y = cen_x * unit_vector_y[0] + cen_y * unit_vector_y[1] + cen_z * unit_vector_y[2]

            # Transform these so that they are in the range [0, sidelength]
            x = x + sidelength_x/2.0
            y = y + sidelength_y/2.0

            #----------------------------------------------------------------------------------#
            # Below this line the code is identical to project_image_omp function. Could clean this up.
            #----------------------------------------------------------------------------------#

            # Position of particle in units of sidelength (0, sidelength)
            x = x*nx/sidelength_x
            y = y*ny/sidelength_y
            h = hvec[ip]*nx/sidelength_x

            if h < 1.:
                h = 1.

            # Index of closest grid point
            ipx = <int> x
            ipy = <int> y

            # Smoothing length as integer
            ih = <int> h + 1

            # Square of smoothing length
            h2 = h*h

            # Find minimum and maximum integers
            ix_min = ipx - ih
            iy_min = ipy - ih
            ix_max = ipx + ih
            iy_max = ipy + ih

            norm = 0.0
            for ix in range(ix_min, ix_max):
                for iy in range(iy_min, iy_max):
                    dx = x - 0.5 - <real_t> ix
                    dy = y - 0.5 - <real_t> iy
                    r2 = dx*dx + dy*dy

                    weight = 1.0 - r2/h2
                    if weight > 0.0:
                        norm = norm + weight

            # Find minimum and maximum integers
            ix_min = max(0, ipx - ih)
            iy_min = max(0, ipy - ih)
            ix_max = min(projection.shape[0], ipx + ih)
            iy_max = min(projection.shape[1], ipy + ih)

            for ix in range(ix_min, ix_max):
                for iy in range(iy_min, iy_max):
                    dx = x - 0.5 - <real_t> ix
                    dy = y - 0.5 - <real_t> iy
                    r2 = dx*dx + dy*dy

                    weight = 1.0 - r2/h2
                    if weight > 0.0:
                        tmp_var[ix, iy, threadnum] = tmp_var[ix, iy, threadnum] \
                                                     + weight*variable[ip]/norm

    # Add up contributions from each thread
    for threadnum in range(numthreads):
        for ix in range(nx):
            for iy in range(ny):
                projection[ix, iy] = projection[ix, iy] + tmp_var[ix, iy, threadnum]

    # Fix to avoid returning a memory-view
    tmp = np.zeros((nx, ny), dtype=np.float64)
    tmp[:, :] = projection[:, :]
    return tmp



from cython.parallel import prange
import numpy as np
cimport numpy as np
cimport libc.math as math

ctypedef fused real_t:
    float
    double
    np.float32_t
    np.float64_t


def get_curvature(real_t[:, :] Bvec, real_t[:, :] Bgradient):

    """
    This function computes the magnetic field line curvature.

    Parameters:
        Bvec (array N, 3): Magnetic field vector
        Bgradient (array N, 3, 3): Magnetic field gradient tensor

    Returns:
        array N: The magnetic field line curvature.
    """
    cdef int ip, ii, jj, kk
    cdef int Np = Bvec.shape[0]
    cdef real_t B2, gradB_ij, gradB_kj, term1, term2

    cdef real_t[:] K2 = np.zeros(Np, dtype=np.float64)

    """
    The code is equivalent to
    Kvec  = np.dot(Bgrad_mat[i,],Bvec[i,]) / B[i]**2
    Kvec -= np.dot(np.dot(Bgrad_mat[i,],Bvec[i,]),Bvec[i,]) * Bvec[i,] / B[i]**4
    K2 = (Kvec**2).sum())
    K = np.sqrt(K2)
    """

    # Bgradient has not been reshaped, \nabla B_ij
    for ip in prange(Np, nogil=True, schedule='static'): #, num_threads=32):
        B2 = 0.0
        for ii in range(3):
            B2 = B2 + Bvec[ip, ii]**2
        for kk in range(3):
            term1 = 0.0
            term2 = 0.0
            for jj in range(3):
                gradB_kj = Bgradient[ip, kk*3+jj]
                term1 = term1 + gradB_kj*Bvec[ip, jj]
                for ii in range(3):
                    gradB_ij = Bgradient[ip, ii*3+jj]
                    term2 = term2 + gradB_ij*Bvec[ip, jj]*Bvec[ip, ii]*Bvec[ip, kk]

            K2[ip] = K2[ip] + (term1/B2 - term2/B2**2.0)**2.0

    # Return a numpy array instead of a view
    tmp = np.zeros(Np, dtype=np.float64)
    tmp[:] = np.sqrt(K2[:])
    return tmp


def get_magnitude_of_vector(real_t[:, :] Bvec):

    """
    Computes the magnitude of a vector, e.g., magnetic field strength.
    """
    cdef int ip, ii
    cdef int Np = Bvec.shape[0]
    cdef real_t B2

    cdef real_t[:] B = np.zeros(Np, dtype=np.float64)

    # Bgradient has not been reshaped, \nabla B_ij
    for ip in prange(Np, nogil=True, schedule='static'):
        B2 = 0.0
        for ii in range(3):
            B2 = B2 + Bvec[ip, ii]**2
        B[ip] = math.sqrt(B2)

    # Return a numpy array instead of a view
    tmp = np.zeros(Np, dtype=np.float64)
    tmp[:] = B[:]
    return tmp

def sum_1d_array_omp(real_t[:] arr, int num_threads):

    """
    Computes the sum of a 1D array using openmp.
    Equivalent to np.sum(arr)
    """
    cdef int ip
    cdef int Np = arr.shape[0]
    cdef real_t the_sum = 0.0

    for ip in prange(Np, nogil=True, schedule='static', num_threads=num_threads):
        the_sum += arr[ip]

    return the_sum

def sum_2d_array_omp(real_t[:, :] arr, int num_threads):

    """
    Computes the sum of an array along the first index using openmp.
    Equivalent to np.sum(arr, axis=0)
    """
    cdef int ip
    cdef int Np = arr.shape[0]
    cdef real_t jj_sum = 0.0

    the_sum = np.zeros(arr.shape[1])
    for jj in range(arr.shape[1]):
        jj_sum = 0.0
        for ip in prange(Np, nogil=True, schedule='static', num_threads=num_threads):
            jj_sum += arr[ip, jj]

        the_sum[jj] = jj_sum

    return the_sum

def sum_arr_times_vector_omp(real_t[:] arr, real_t[:, :] vector, int num_threads):

    """
    Computes the sum of an array times a vector.
    e.g. sum_i M_i vec{v}_i for calculating the center of mass.

    Equivalent to np.sum(arr[:, None] * vector, axis=0)
    """
    cdef int ip
    cdef int Np = arr.shape[0]
    cdef real_t the_sum_x = 0.0
    cdef real_t the_sum_y = 0.0
    cdef real_t the_sum_z = 0.0

    for ip in prange(Np, nogil=True, schedule='static', num_threads=num_threads):
        the_sum_x += arr[ip] * vector[ip, 0]
        the_sum_y += arr[ip] * vector[ip, 1]
        the_sum_z += arr[ip] * vector[ip, 2]

    return np.array([the_sum_x, the_sum_y, the_sum_z])

def sum_arr_times_vector_cross_product(real_t[:] mass, real_t[:, :] coord, real_t[:, :] velocity,
                                       real_t[:] center, int num_threads):

    """
    This code calculates sum_i (mass_i (coord_ij - center) x velocity_ij).

    That is, it returns the total angular momentum vector.
    """
    cdef int ip
    cdef int Np = mass.shape[0]
    cdef real_t the_sum_x = 0.0
    cdef real_t the_sum_y = 0.0
    cdef real_t the_sum_z = 0.0
    cdef real_t vx, vy, vz, rx, ry, rz

    for ip in prange(Np, nogil=True, schedule='static', num_threads=num_threads):
        vx = velocity[ip, 0]
        vy = velocity[ip, 1]
        vz = velocity[ip, 2]
        rx = coord[ip, 0] - center[0]
        ry = coord[ip, 1] - center[1]
        rz = coord[ip, 2] - center[2]
        the_sum_x += mass[ip] * (ry*vz - rz*vy)
        the_sum_y += mass[ip] * (rz*vx - rx*vz)
        the_sum_z += mass[ip] * (rx*vy - ry*vx)

    return np.array([the_sum_x, the_sum_y, the_sum_z])



import numpy as np
from .image_creator import ImageCreator
from .. import util
from .. import settings
from .. import units

import cupy as cp
from numba import cuda


@cuda.jit(device=True, inline=True)
def get_norm_and_index(xp, yp, hp, extra):
    ipx = int(xp)
    ipy = int(yp)

    h2 = hp**2

    norm = 0.0
    for ix in range(-extra, extra):
        for iy in range(-extra, extra):
            dx = xp - 0.5 - ix - ipx
            dy = yp - 0.5 - iy - ipy
            r2 = dx * dx + dy * dy

            weight = 1.0 - r2 / h2

            if weight < 0.0:
                weight = 0.0

            norm += weight

    return norm, ipx, ipy


@cuda.jit(device=True, inline=True)
def get_weight(ix, iy, xp, yp, hp, varp):
    ipx = int(xp)
    ipy = int(yp)

    h2 = hp**2

    dx = xp - 0.5 - ix - ipx
    dy = yp - 0.5 - iy - ipy
    r2 = dx * dx + dy * dy

    weight = 1.0 - r2 / h2
    # return weight
    if weight < 0.0:
        return 0.0
    else:
        return weight * varp


@cuda.jit(max_registers=64)
def deposit(nx, ny, x, y, z, hsml, variable, image1, image2, image4, image8, image16,
            widths, center, unit_vector_x, unit_vector_y, unit_vector_z, extra):
    # threadindex
    ip = cuda.grid(1)

    # particle properties
    hp = hsml[ip]
    xp = x[ip]
    yp = y[ip]
    zp = z[ip]
    varp = variable[ip]

    # Center of projection
    xc, yc, zc = center

    # Centered coordinates
    cen_x = xp - xc
    cen_y = yp - yc
    cen_z = zp - zc

    # Do projections along unit vectors
    # Projection of coordinate along perp_vector1
    xp = cen_x * unit_vector_x[0] + cen_y * \
        unit_vector_x[1] + cen_z * unit_vector_x[2]
    # Projection of coordinate along perp_vector2
    yp = cen_x * unit_vector_y[0] + cen_y * \
        unit_vector_y[1] + cen_z * unit_vector_y[2]
    # Projection of coordinate along normal_vector
    zp = cen_x * unit_vector_z[0] + cen_y * \
        unit_vector_z[1] + cen_z * unit_vector_z[2]

    # For oriented images the projection is rotated to be
    # in a coordinate system where the projection direction is
    # z. The sidelength are the x-axis and y-axis of the image,
    # not the x and y dimensions of the simulation
    sidelength_x, sidelength_y, sidelength_z = widths

    # Check if this cell/particle is inside domain
    inside_domain = False
    if (xp < sidelength_x / 2.0) and (xp > -sidelength_x / 2.0):
        if (yp < sidelength_y / 2.0) and (yp > -sidelength_y / 2.0):
            if (zp < sidelength_z / 2.0):
                if (zp > -sidelength_z / 2.0):
                    inside_domain = True

    if inside_domain:

        # Transform these so that they are in the range [0, sidelength]
        xp = xp + sidelength_x / 2.0
        yp = yp + sidelength_y / 2.0

        # Image cell size on finest grid
        dx = sidelength_x / nx

        # Position of particle in units of sidelength (0, sidelength)
        xp = xp / dx
        yp = yp / dx
        hp = hp / dx

        # Use a minimum cell size for projection
        if hp < 1.0:
            hp = 1.0

        # Split into the 5 grids at different resolutions
        if hp < extra:
            im = 1
        elif hp < 2 * extra:
            im = 2
        elif hp < 4 * extra:
            im = 4
        elif hp < 8 * extra:
            im = 8
        else:
            im = 16

        # Change cell size depending on the image cell size
        xp /= im
        yp /= im
        hp /= im

        # Find norm of particle (ensures that all its 'mass' is deposited)
        norm, ipx, ipy = get_norm_and_index(xp, yp, hp, extra)

        # Loop over the stencil
        for d_ix in range(-extra, extra):
            for d_iy in range(-extra, extra):
                ix = ipx + d_ix
                iy = ipy + d_iy
                if ix >= 0 and ix < nx // im and iy >= 0 and iy < ny // im:
                    weight = get_weight(d_ix, d_iy, xp, yp, hp, varp)
                    if im == 1:
                        cuda.atomic.add(image1, (ix, iy), weight / norm)
                    elif im == 2:
                        cuda.atomic.add(image2, (ix, iy), weight / norm)
                    elif im == 4:
                        cuda.atomic.add(image4, (ix, iy), weight / norm)
                    elif im == 8:
                        cuda.atomic.add(image8, (ix, iy), weight / norm)
                    elif im == 16:
                        cuda.atomic.add(image16, (ix, iy), weight / norm)


class GpuSphProjector(ImageCreator):
    """
    A class that allows creating an image of a given variable by projecting
    it onto a 2D plane.

    This GPU implementation of SPH-like projection splits the particles
    to be deposited into five nested grids. The current implementation
    gives visual artifacts when the number of pixels is chosen to be too
    large compared to the gas cell sizes (i.e. when trying
    to make a high-resolution image of a low resolution simulation).
    This projector class is therefore for the moment only recommended
    for interactive exploration or for creating movies.

    The input pixel size needs to be a power of 2 in the horizontal direction.
    """

    def __init__(self, snap, center, widths, direction,
                 npix=512, parttype=0, nvol=8, threadsperblock=8,
                 do_pre_selection=False):
        """
        Initialize the Projector class.

        Parameters
        ----------
        snap : Snapshot
            A snapshot object of Snapshot class from paicos package.

        center : numpy array
            Center of the region on which projection is to be done, e.g.
            center = [x_c, y_c, z_c].

        widths : numpy array
            Widths of the region on which projection is to be done,
            e.g.m widths=[width_x, width_y, width_z].

        direction : str
            Direction of the projection, e.g. 'x', 'y' or 'z'.

        npix : int, optional
            Number of pixels in the horizontal direction of the image,
            by default 512.

        parttype : int, optional
            Number of the particle type to project, by default gas (PartType 0).

        nvol : int, optional
            Integer used to determine the smoothing length, by default 8

        """

        # call the superclass constructor to initialize the ImageCreator class
        super().__init__(snap, center, widths, direction, npix=npix, parttype=parttype)
        err_msg = 'Pixel sizes in both directions must be divisible by 32'
        assert self.npix_width % 32 == 0, err_msg
        assert self.npix_height % 32 == 0, err_msg

        parttype = self.parttype

        self.threadsperblock = threadsperblock

        self.do_pre_selection = do_pre_selection

        # nvol is an integer that determines the smoothing length
        self.nvol = nvol

        # Calculate the smoothing length
        avail_list = (list(snap.keys()) + snap._auto_list)
        if f'{parttype}_Volume' in avail_list:
            self.hsml = np.cbrt(nvol * (self.snap[f"{parttype}_Volume"])
                                / (4.0 * np.pi / 3.0))
        elif f'{parttype}_SubfindHsml' in avail_list:
            self.hsml = self.snap[f'{parttype}_SubfindHsml']
        else:
            raise RuntimeError(
                'There is no smoothing length or volume for the projector')

        self.pos = self.snap[f'{self.parttype}_Coordinates']

        if settings.use_units:
            self.hsml = self.hsml.to(self.pos.unit)

        # TODO: add split into GPU and CPU based on cell sizes here.
        # Perhaps make the CPU do everything above grid8.

        self._sph_kernel_radius = 4

        # Call selection
        self.has_do_region_selection_been_called = False
        self._do_region_selection()
        self.has_do_region_selection_been_called = True

    def _do_region_selection(self):

        self.do_unit_consistency_check()

        center = self.center
        widths = self.widths
        snap = self.snap
        parttype = self.parttype

        if self.has_do_region_selection_been_called:
            if self.do_pre_selection:
                import warnings
                err_msg = ("It looks like you are changing projector "
                           + " properties after the fact, i.e. changing widths "
                           + "center, orientation, resolution etc. This might be "
                           + "slow with the option do_pre_selection, which "
                           + "you have turned on. If your GPU has enough memory "
                           + "then it is probably better to set do_pre_selection "
                           + "to False.")
                warnings.warn(err_msg)

        # Send subset of snapshot to GPU
        if self.do_pre_selection:
            # get the index of the region of projection
            if self.direction != 'orientation':
                get_index = util.get_index_of_cubic_region
                self.index = get_index(self.snap[f"{parttype}_Coordinates"],
                                       center, widths, snap.box)
            else:
                get_index = util.get_index_of_rotated_cubic_region
                self.index = get_index(snap[f"{parttype}_Coordinates"],
                                       center, widths, snap.box,
                                       self.orientation)

            self.hsml = self.hsml[self.index]
            self.pos = self.pos[self.index]

            self._send_data_to_gpu()
            self._send_small_data_to_gpu()
        # Send entirety of snapshot to GPU (if we have not already
        # done so). Always send small data with change in resolution etc
        else:
            if not self.has_do_region_selection_been_called:
                self._send_data_to_gpu()
            self._send_small_data_to_gpu()

    def _send_data_to_gpu(self):
        self.gpu_variables = {}
        if settings.use_units:
            self.gpu_variables['x'] = cp.array(self.pos[:, 0].value)
            self.gpu_variables['y'] = cp.array(self.pos[:, 1].value)
            self.gpu_variables['z'] = cp.array(self.pos[:, 2].value)
            self.gpu_variables['hsml'] = cp.array(self.hsml.value)
        else:
            self.gpu_variables['x'] = cp.array(self.pos[:, 0])
            self.gpu_variables['y'] = cp.array(self.pos[:, 1])
            self.gpu_variables['z'] = cp.array(self.pos[:, 2])
            self.gpu_variables['hsml'] = cp.array(self.hsml)

        self._send_small_data_to_gpu()

    def _send_small_data_to_gpu(self):
        unit_vectors = self.orientation.cartesian_unit_vectors
        self.gpu_variables['unit_vector_x'] = cp.array(unit_vectors['x'])
        self.gpu_variables['unit_vector_y'] = cp.array(unit_vectors['y'])
        self.gpu_variables['unit_vector_z'] = cp.array(unit_vectors['z'])

        if settings.use_units:
            self.gpu_variables['widths'] = cp.array(self.widths.value)
            self.gpu_variables['center'] = cp.array(self.center.value)
        else:
            self.gpu_variables['widths'] = cp.array(self.widths)
            self.gpu_variables['center'] = cp.array(self.center)

    def _increase_image_resolution(self, image, factor):
        """
        Increase the number of pixes without changing the total 'mass'
        of the image
        """
        if factor == 1:
            return image
        repeats = factor
        new_image = cp.repeat(cp.repeat(image, repeats, axis=0),
                              repeats, axis=1)
        return new_image / factor**2

    def _get_full_image_as_numpy(self, image1, image2, image4, image8, image16):
        image8 += self._increase_image_resolution(image16, 2)
        image4 += self._increase_image_resolution(image8, 2)
        image2 += self._increase_image_resolution(image4, 2)
        image1 += self._increase_image_resolution(image2, 2)
        return cp.asnumpy(image1)

    def _gpu_project(self, variable_str):
        """
        Private method for projecting using cuda code
        """

        unit_vector_x = self.gpu_variables['unit_vector_x']
        unit_vector_y = self.gpu_variables['unit_vector_y']
        unit_vector_z = self.gpu_variables['unit_vector_z']
        x = self.gpu_variables['x']
        y = self.gpu_variables['y']
        z = self.gpu_variables['z']
        widths = self.gpu_variables['widths']
        center = self.gpu_variables['center']
        hsml = self.gpu_variables['hsml']
        nx = self.npix_width
        ny = self.npix_height
        variable = self.gpu_variables[variable_str]

        threadsperblock = self.threadsperblock
        blockspergrid = (x.size + (threadsperblock - 1)) // threadsperblock
        image1 = cp.zeros((nx, ny))
        image2 = cp.zeros((nx // 2, ny // 2))
        image4 = cp.zeros((nx // 4, ny // 4))
        image8 = cp.zeros((nx // 8, ny // 8))
        image16 = cp.zeros((nx // 16, ny // 16))

        deposit[blockspergrid, threadsperblock](nx, ny, x, y, z, hsml, variable,
                                                image1, image2, image4, image8, image16,
                                                widths, center,
                                                unit_vector_x,
                                                unit_vector_y,
                                                unit_vector_z, self._sph_kernel_radius)

        projection = self._get_full_image_as_numpy(
            image1, image2, image4, image8, image16)
        return projection

    def project_variable(self, variable):
        """
        projects a given variable onto a 2D plane.

        Parameters
        ----------
        variable : str, function, numpy array
            variable, it can be passed as string or an array

        Returns
        -------
        numpy array
            The image of the projected variable
        """

        self.do_unit_consistency_check()

        # This calls _do_region_selection if resolution, Orientation,
        # widths or center changed
        self._check_if_properties_changed()

        if isinstance(variable, str):
            variable_str = str(variable)
            err_msg = 'projector uses a different parttype'
            assert int(variable[0]) == self.parttype, err_msg
            variable = self.snap[variable]
        else:
            variable_str = 'projection_variable'
            if not isinstance(variable, np.ndarray):
                raise RuntimeError('Unexpected type for variable')

        assert len(variable.shape) == 1, 'only scalars can be projected'

        # Select same part of array that the projector has selected
        if self.do_pre_selection:
            variable = variable[self.index]

        if variable_str in self.gpu_variables and variable_str != 'projection_variable':
            pass
        else:
            # Send variable to gpu
            if settings.use_units:
                self.gpu_variables[variable_str] = cp.array(variable.value)
            else:
                self.gpu_variables[variable_str] = cp.array(variable)

        # Do the projection
        projection = self._gpu_project(variable_str)

        # Transpose
        projection = projection.T

        assert projection.shape[0] == self.npix_height
        assert projection.shape[1] == self.npix_width

        if isinstance(variable, units.PaicosQuantity):
            projection = projection * variable.unit_quantity

        return projection / self.area_per_pixel

    def __del__(self):
        """
        Clean up like this? Not sure it is needed...
        """
        del self.gpu_variables
        # for key in list(self.gpu_variables.keys):
        #     del self.gpu_variables[key]
        cp._default_memory_pool.free_all_blocks()



import setuptools
from setuptools import setup
from setuptools import Extension
from Cython.Build import cythonize
from Cython.Compiler import Options
import Cython
import sys
import os
import numpy
import glob


# On MacOs, you will need to explicitly set the compiler to one that features openmp.
# Either by uncommenting below or by compiling with e.g. 'CC=gcc-14 make'
# if sys.platform == 'darwin':
#     os.environ['CC'] = 'gcc-13'
#     os.environ['CXX'] = 'g++-13'

Options.annotate = False
compiler_directives = {"boundscheck": False, "cdivision": True,
                       "wraparound": False, 'language_level': "3"}


include_dirs = ['paicos/cython/', numpy.get_include()]
extra_compile_args = ['-fopenmp', "-DNPY_NO_DEPRECATED_API=NPY_1_7_API_VERSION",
                      "-Wno-unused-function"]
extra_link_args = ['-fopenmp']
ext_modules = [
    Extension(
        name='paicos.cython.get_index_of_region',
        sources=['paicos/cython/get_index_of_region.pyx'],
        include_dirs=include_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args
    ),
    Extension(
        name='paicos.cython.histogram',
        sources=['paicos/cython/histogram.pyx'],
        include_dirs=include_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args
    ),
    Extension(
        name='paicos.cython.sph_projectors',
        sources=['paicos/cython/sph_projectors.pyx'],
        include_dirs=include_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args
    ),
    Extension(
        name='paicos.cython.openmp_info',
        sources=['paicos/cython/openmp_info.pyx'],
        include_dirs=include_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args
    ),
    Extension(
        name='paicos.cython.get_derived_variables',
        sources=['paicos/cython/get_derived_variables.pyx'],
        include_dirs=include_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args
    ),
    Extension(
        name='paicos.trees.bvh_tree',
        sources=['paicos/trees/bvh_tree.pyx'],
        include_dirs=include_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args
    )

]

install_requires = ['scipy',
                    'numpy',
                    'h5py',
                    'astropy',
                    'numba',
                    'matplotlib']


cur = os.path.abspath(__file__).replace('setup.py', '')
with open(cur + 'README.md') as f:
    long_description = f.read()

with open(cur + 'dev_requirements.txt') as f:
    dev_requirements = f.read().strip().split('\n')

setup(
    name='paicos',
    version='0.1.12',
    description=('An object-oriented Python package for analysis of '
                 + '(cosmological) simulations performed with Arepo.'),
    url='https://github.com/tberlok/paicos',
    author='Thomas Berlok',
    author_email='tberlok@gmail.com',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='BSD 3-clause',
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    extras_require={'dev': dev_requirements},
    package_data={'paicos/cython': ['*.c', '*.so']},
    classifiers=['Programming Language :: Python :: 3'],
    ext_modules=cythonize(ext_modules,
                          compiler_directives=compiler_directives),
    cmdclass={'build_ext': Cython.Build.build_ext})



"""
This defines a class for saving images as hdf5 files in a systematic way.
"""

import h5py
import numpy as np
from .. import util
from .paicos_writer import PaicosWriter
from .. import settings
from .. import units
from ..orientation import Orientation


class ArepoImage(PaicosWriter):
    """
    A derived data format for Arepo snapshots.

    A common task is to reduce Arepo snapshots to 2D arrays (either slices or
    projections) which can plotted as images. The computation of the 2D arrays
    can be time-consuming for high-resolution simulations with large
    snapshots. The purpose of this class is to define a derived data format
    which can be used to store images for later plotting with matplotlib.
    """

    def __init__(self, image_creator, basedir, basename="projection", mode='w'):
        """
        Initialize an HDF5 file for storing an image.

        This class is intended to be used with images created using a Paicos
        Projector or Slicer object, which can be passed as the `image_creator`
        input argument.

        The image data, including 'center', 'widths', 'extent', and
        'direction', will be extracted from the `image_creator` object and
        stored in the HDF5 file. The HDF5 file is created in the `basedir`
        folder, with the name `basename_{:03d}`.

        The `mode` argument controls whether the file is opened in write mode
        ('w') or amend mode ('a'). If `mode` is 'w', the file will be created
        at `self.tmp_filename` and the image information will be written to
        the file. Setting the mode to amend mode, 'a', allows to add new images
        to an already existing file.

        Parameters
        ----------

            image_creator : object
                A Paicos Projector or Slicer object used to create the image.

            basedir (file: path
                The folder where the image file should be saved.

            basename : string
                The base name for the image file, which will be in
                the format ``basename_{:03d}``. (default: "projection").

            mode : string
                The mode to open the file in, either 'w' for write mode
                or 'a' for append mode. (default: 'w').

        Methods
        -------

            finalize :
                Changes the filename from a temporary one to the final one,
                e.g., from self.tmp_filename to self.filename.
        """

        self.center = image_creator.center
        self.widths = image_creator.widths
        self.extent = image_creator.extent
        self.direction = image_creator.direction
        self.orientation = image_creator.orientation

        # This creates an image at self.tmp_filename (if mode='w')
        super().__init__(image_creator.snap, basedir, basename=basename,
                         mode=mode)

        # Create image file and write information about image
        if self.mode == 'w':
            with h5py.File(self.tmp_filename, 'r+') as file:
                util.save_dataset(file, 'center', self.center, group='image_info')
                util.save_dataset(file, 'widths', self.widths, group='image_info')
                util.save_dataset(file, 'extent', self.extent, group='image_info')
                file['image_info'].attrs['direction'] = self.direction
                file['image_info'].attrs['image_creator'] = str(image_creator)

                if self.direction == 'orientation':
                    file['image_info'].attrs['normal_vector'] = self.orientation.normal_vector
                    file['image_info'].attrs['perp_vector1'] = self.orientation.perp_vector1

    def save_image(self, name, data):
        """
        This method saves a 2D image to the hdf5 file.
        """
        self.write_data(name, data)

    def _perform_extra_consistency_checks(self):
        """
        Perform extra consistency checks on the HDF5 file to ensure the
        current values match the values stored in the HDF5 file that we intend
        to amend.

        The function opens the HDF5 file, loads the 'center' and 'widths'
        datasets from the 'image_info' group, and compares them to the
        corresponding values stored in memory. The 'direction' attribute of
        the 'image_info' group is also checked for consistency.
        """
        with h5py.File(self.filename, 'r') as f:
            center = util.load_dataset(f, 'center', group='image_info')
            if settings.use_units:
                assert center.unit == self.center.unit
                np.testing.assert_array_equal(center.value, self.center.value)
            else:
                np.testing.assert_array_equal(center, self.center)

            widths = util.load_dataset(f, 'widths', group='image_info')
            if settings.use_units:
                assert widths.unit == self.widths.unit
                np.testing.assert_array_equal(widths.value, self.widths.value)
            else:
                np.testing.assert_array_equal(widths, self.widths)

            assert f['image_info'].attrs['direction'] == self.direction



"""
This is where we import all the various default functions for getting derived variables.
"""
from . import derived_variables_gas

user_functions = {}

default_functions = {}

# Add all gas functions to the defaults
for func_name, func in derived_variables_gas.functions.items():
    default_functions[func_name] = func
